#!/usr/bin/env python

import os, sys, datetime, logging, json, time

from kafka import KafkaProducer
from json import loads
from argparse import ArgumentParser


def _exit(message = None, rc=0):
  if (message): printMsg(message)
  printMsg("Exiting [rc={}]".format(rc))
  sys.exit(rc)


def timeFormat():
  return '%b.%d.%Y-%H.%M.%S.%f'


def printMsg(message):
  print ("[{}] [{}] {}".format(datetime.datetime.now().strftime(timeFormat()), GNode, message))


class MyKafkaProducer:
    kafka_service = os.environ.get('KAFKA_SERVER')
    kafka_port = os.environ.get('KAFKA_PORT')

    if not kafka_service: kafka_service = "localhost"
    if not kafka_port: kafka_port = "9092"

    server = kafka_service + ':' + kafka_port

    def __init__(self):
        self.producer = KafkaProducer(bootstrap_servers=[self.server],
                        value_serializer=lambda x:
                        json.dumps(x).encode('utf-8'))

    def send_data(self, topic, data):
        json = self.__read_json_data(data)
        self.producer.send(topic, value=json)
        self.producer.flush()
        printMsg("Json produced on topic '{}': {}".format(topic, json))

    def __read_json_data(self,jsonData):
        with open(jsonData,'r') as schema_file:
            schema_data = schema_file.read()
        schema = json.loads(schema_data)
        return schema

    def set_producer_topic(self, kafka_topic):
        self.kafka_topic = kafka_topic

    def close(self):
        self.producer.close()

class MyProducer:
    #def __init__(self):
    producer = MyKafkaProducer()

    def send_data(self, topic, data):
        self.producer.send_data(topic, data)

    def close(self):
        self.producer.close()


def parse_arguments():

  parser = ArgumentParser(description='Kafka Producer Simulator')
  parser.add_argument('-n', '--node', help='Node name (spaces are removed)', required=True)
  parser.add_argument('-t', '--topic', help='Kafka topic', required=True)
  parser.add_argument('-w', '--delay', help='Delay (in milliseconds) before producing', required=False)
  parser.add_argument('-f', '--file', help='Json file to produce', required=True)
  parser.add_argument('-d', '--debug', help='Debug kafka', required=False, action='store_true')

  arguments = parser.parse_args()

  return arguments


#####################
#      M A I N      #
#####################

if __name__ == "__main__":

  arguments = parse_arguments()

  GNode = arguments.node.replace(" ", "")
  topic = arguments.topic
  delay = arguments.delay
  jfile = arguments.file
  debug = arguments.debug

  printMsg("Starting")

  if (debug):
    now = datetime.datetime.now().strftime(timeFormat())
    logfile='{}-Producer-{}.log'.format(GNode, now)
    printMsg("Logging kafka in: '{}'".format(logfile))
    logging.basicConfig(filename=logfile, format='%(asctime)s - %(message)s', datefmt=timeFormat(), level=logging.DEBUG)

  if not os.path.exists(jfile):
    _exit("Invalid provided file '{}'".format(jfile), 1)

  if (delay):
    try:
      delay = int(delay)
    except:
      _exit("Invalid provided delay '{}'. Must be an integer value !".format(str(delay)))

    printMsg("Delay before producing: {} milliseconds ...".format(delay))
    time.sleep(delay / 1000)
    printMsg("Delay completed !".format(delay))

  producer = None
  try:
    producer = MyProducer()
    producer.send_data(topic, jfile)
    producer.close()
    _exit()

  except Exception as e:
    producer.close()
    _exit(e, 1)
