// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_Average_hpp
#define anna_core_util_Average_hpp

#include <anna/core/util/Millisecond.hpp>
#include <anna/core//util/DelayMeter.hpp>
#include <anna/core/functions.hpp>

namespace anna {

/**
   Template para calcular la media de una sucesion de valores numericos.

   Internamente trabaja con Microsegundos porque los mtodos que los proporcionan son
   ms eficientes.
*/
template <class T> class Average {
public:
  /**
     Constructor.
     \param name Nombre logico de esta instancia.
     \param measure Unidad de medida. Solo se usa a efecto de salida de datos.
  */
  Average(const char* name, const char* measure = NULL) :
    a_name(name), a_accumulator(0), a_n(0), a_measure(measure) {;}

  /**
     Devuelve el indicador de validez de esta media.
     \return \em true Si la media no tiene ningun valor o \em false en caso contrario.
  */
  bool isEmpty() const throw()  { return (a_n == 0); }

  /**
     Devuelve \em true si el valor de la media es cero, bien por no tener valores o
     bien porque todos ellos eran cero.
     \return \em true el valor de la media es cero o \em false en otro caso.
  */
  bool isZero() const throw() { return a_n == 0 || a_accumulator == 0; }

  /**
     Devuelve el numero de elementos de contiene la media.
     \return el numero de elementos de contiene la media.
  */
  int size() const throw() { return a_n; }

  /**
     Devuelve el valor acumulado.
     \return el valor acumulado.
  */
  T getAccumulator() const throw() { return a_accumulator; }

  /**
     Devuelve la media de la sucesion de valores numericos asociados a esta.
     \return La media de la sucesion de valores numericos asociados a esta.
     \warning Antes de invocar a este operador hay que verificar que #isEmpty devuelve \em false.
  */
  T value() const throw() { return (isEmpty() == true) ? T(0) : (a_accumulator / a_n);  }

  /**
     Inicializa el valor de la media.
  */
  void clear() throw() {  a_accumulator = 0; a_n = 0; }

  /**
   * Establece manualmente el valor de la estadstica.
   * \param value Valor que tomar el acumulador de este instancia.
   * \param _n valor que tomar el conteador de esta instancia.
   */
  void setValue(const T& value, const unsigned int _n) throw() {
    a_accumulator = value;
    a_n = _n;
  }

  /**
     Devuelve la media de la sucesion de valores numericos asociados a esta.
     \return La media de la sucesion de valores numericos asociados a esta.
     \warning Antes de invocar a este operador hay que verificar que #isEmpty devuelve \em false.
  */
  operator T() const throw() { return value(); }

  /**
     Inicializa el valor de esta media.
     \param value Valor con el que iniciar la media.
     \return La referencia a esta instancia.
  */
  Average<T>& operator = (const T value)
  throw() {
    a_accumulator = value;
    a_n = 1;
    return *this;
  }

  /**
     Contructor copia.
     \param other Objeto del que obtener los valores.
     \return La referencia a esta instancia.
  */
  Average<T>& operator = (const Average<T>& other)
  throw() {
    a_accumulator = other.a_accumulator;
    a_n = other.a_n;
    return *this;
  }

  /**
     Incrementa la media con el valor recibido.
     \param v Valor con el que incrementar la media.
     \return La referencia a esta instancia.
  */
  Average<T>& operator += (const T& v)
  throw() {
    const T backup(a_accumulator);

    if((a_accumulator += v) < backup) {
      a_accumulator = v;
      a_n = 1;
    } else
      a_n ++;

    return *this;
  }

  /**
     Decrementa la media con el valor recibido.
     \param v Valor con el que incrementar la media.
     \return La referencia a esta instancia.
  */
  Average<T>& operator -= (const T v)
  throw() {
    if(a_accumulator > v && a_n > 1) {
      a_accumulator -= v;
      a_n --;
    } else {
      a_accumulator = 0;
      a_n = 0;
    }

    return *this;
  }

  /**
     Devuelve una cadena con la informacion referente a esta clase.
     \return Una cadena con la informacion referente a esta clase.
  */
  std::string asString() const
  throw() {
    std::string msg(a_name);
    msg += " { Accumulate: ";
    msg += functions::asString(a_accumulator);

    if(a_measure != NULL)
      msg += " " + std::string(a_measure);

    msg += functions::asString(" | N: %u | avg: ", a_n);
    msg += functions::asString(value());

    if(a_measure != NULL)
      msg += " " + std::string(a_measure);

    return msg += " }";
  }

private:
  const char* a_name;
  const char* a_measure;
  T a_accumulator;
  unsigned int a_n;
};

}

#endif



