// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_LRUMap_hpp
#define anna_core_util_LRUMap_hpp

#include <map>

#include <anna/config/defines.hpp>

namespace anna {

/**
   Patrn que permite mantener una lista de N elementos de pares (Clave, Valor).

   Es capaz de realizar una bsqueda indexada de la clave para obtener el valor,
   pero adems es capaz de mantener estable el nmero de elementos contenidos
   en sta clase.

   Si se alcanza el nmero mximo de elementos indicados en el constructor, si se requiere
   incluir un nuevo objeto, primero liberar el objeto que lleve ms tiempo si ser usado.

   \param K clase que actuar como clave del mapa. Requiere el contructor de copia, y los operadores de comparacin == y <.
   \param V clase que actuar como dato asociado a la clave. Requiere el constructor copia y el operador copia.
*/
template <typename K, typename V> class LRUMap {
  typedef typename std::pair <V, anna::Millisecond> timed_value;
  typedef typename std::map <K, timed_value> container;
  typedef typename container::value_type value_type;

public:
  typedef typename container::iterator iterator;
  typedef typename container::const_iterator const_iterator;

  /**
     Constructor.
     \param name Nombre logico de esta instancia.
     \param measure Unidad de medida. Solo se usa a efecto de salida de datos.
  */
  LRUMap(const char* name, const int maxSize) : a_name(name), a_maxSize(maxSize) { ;}

  /**
   * Destructor.
   */
  ~LRUMap() { clear(); }

  /**
     Devuelve el indicador de validez de esta media.
     \return \em true Si la media no tiene ningun valor o \em false en caso contrario.
  */
  bool isEmpty() const throw() { return a_container.size() == 0; }

  /**
   * Devuelve el nmero de elementos contenidos en este contenedor.
   * \return  el nmero de elementos contenidos en este contenedor.
   */
  int size() const throw() { return a_container.size(); }

  /**
   * Devuelve el puntero al valor asociado a la clave recibida como parmetro.
   * Puede ser NULL si la pareja (K,V) no est registrada en el contenedor.
   *
   * \param key Clave de la que obtener el valor asociado.
   *
   * \return El puntero al valor asociado a la clave recibida como parmetro.
   * */
  V* find(const K& key) throw() {
    iterator ii =  a_container.find(key);

    if(ii == end())
      return NULL;

    /*
     * Recupera el valor asociado a la clave K y actualiza el momento de acceso.
     */
    millisecond(ii) = functions::millisecond();
    return &value(ii);
  }

  /**
   * Establece el valor de pareja (K,V). Si no existe se crea y si existe se sobre escribe el
   * valor asociado a V, y se actualiza su tiempo de acceso.
   *
   * \param key Clave de la pareja (K,V).
   * \param v Valor asociado a la clave.
   */
  void add(const K& key, const V& v) throw() {
    iterator ii = a_container.find(key);

    // Sobreescribe el valor asociado a K y actualiza su tiempo de acceso.
    if(ii != end()) {
      value(ii) = v;
      millisecond(ii) = functions::millisecond();
      return;
    }

    // Si no se ha alcanzado el mximo, slo hay que insertar el nuevo (K,V)
    if(size() < a_maxSize) {
      timed_value tvalue(v, functions::millisecond());
      a_container.insert(value_type(key, tvalue));
      return;
    }

    // Se ha alcanzado el mximo, hay que buscar el que haga ms tiempo que no se
    // accede => el que tenga el menor 'Time' asociado.
    iterator minii = begin();
    Millisecond minTime = millisecond(minii);

    for(iterator ii = begin(), maxii = end(); ii != maxii; ii ++) {
      if(millisecond(ii) < minTime)
        minTime = millisecond(minii = ii);
    }

    /*
     * El map no permite sobre-escribir la clave => por eso tenemos que borrar el nodo
     * ms antiguo y crear otro nuevo.
     */
    a_container.erase(minii);
    timed_value tvalue(v, functions::millisecond());
    a_container.insert(value_type(key, tvalue));
  }

  /**
     Vaca este contenedor.
  */
  void clear() throw() { a_container.clear(); }

  /**
   * Devuelve un iterator al primer elemento del contenedor, teniendo que la ordenacin de los
   * pares (K,V) se har en base a K.
   * \return El primer elemento del contenedor.
  */
  iterator begin() throw() { return a_container.begin(); }

  /**
   * Devuelve un iterator al primer elemento del contenedor, teniendo que la ordenacin de los
   * pares (K,V) se har en base a K.
   * \return El primer elemento del contenedor.
  */
  const_iterator begin() const throw() { return a_container.begin(); }

  /**
   * Devuelve un iterator al final del contenedor, teniendo que la ordenacin de los  pares (K,V) se har en
   * base a K.
   * \return El primer elemento del contenedor.
   */
  iterator end() throw() { return a_container.end(); }

  /**
   * Devuelve un iterator al final del contenedor, teniendo que la ordenacin de los  pares (K,V) se har en
   * base a K.
   * \return El primer elemento del contenedor.
   */
  const_iterator end() const throw() { return a_container.end(); }

  /**
   * Devuelve la clave asociada al iterador recibido como parmetro.
   * \param ii Iterador que debe estar comprendido entre  [#begin (), #end).
   * \return la clave asociada al iterador recibido como parmetro.
   * \warning Los acceso mediante iterador no actualiza el tiempo de acceso a la pareja (K,V).
   */
  static K key(iterator& ii) throw() { return ii->first;  }

  /**
   * Devuelve el valor asociado al iterador recibido como parmetro.
   * \param ii Iterador que debe estar comprendido entre  [#begin (), #end).
   * \return el valor asociado al iterador recibido como parmetro.
   * \warning Los acceso mediante iterador no actualiza el tiempo de acceso a la pareja (K,V).
   */
  static V& value(iterator& ii) throw() {
    timed_value* v = &ii->second;
    return v->first;
  }

  /**
   * Devuelve el tiempo de acceso asociado al iterador recibido como parmetro.
   * \param ii Iterador que debe estar comprendido entre  [#begin (), #end).
   * \return el tiempo de acceso asociado al iterador recibido como parmetro.
   */
  static Millisecond& millisecond(iterator& ii) throw() {
    timed_value* v = &ii->second;
    return v->second;
  }

  /**
   * Devuelve la clave asociada al iterador recibido como parmetro.
   * \param ii Iterador que debe estar comprendido entre  [#begin (), #end).
   * \return la clave asociada al iterador recibido como parmetro.
   * \warning Los acceso mediante iterador no actualiza el tiempo de acceso a la pareja (K,V).
   */
  static K key(const_iterator& ii) throw() { return ii->first;  }

  /**
   * Devuelve el valor asociado al iterador recibido como parmetro.
   * \param ii Iterador que debe estar comprendido entre  [#begin (), #end).
   * \return el valor asociado al iterador recibido como parmetro.
   * \warning Los acceso mediante iterador no actualiza el tiempo de acceso a la pareja (K,V).
   */
  static const V& value(const_iterator& ii) throw() {
    const timed_value* v = &ii->second;
    return v->first;
  }

  /**
   * Devuelve el tiempo de acceso asociado al iterador recibido como parmetro.
   * \param ii Iterador que debe estar comprendido entre  [#begin (), #end).
   * \return el tiempo de acceso asociado al iterador recibido como parmetro.
   */
  static Millisecond millisecond(const_iterator& ii) throw() {
    const timed_value* v = &ii->second;
    return v->second;
  }

  /**
     Devuelve una cadena con la informacion referente a esta clase.
     \return Una cadena con la informacion referente a esta clase.
  */
  std::string asString() const
  throw() {
    std::string msg("LRUMap { Name: ");
    msg += a_name;
    msg += functions::asText(" | N: ", a_maxSize);
    msg += functions::asText(" | n: ", size());
    return msg += " }";
  }

private:
  const char* a_name;
  const int a_maxSize;
  container a_container;
};

}

#endif



